// Copyright 2020 The gVisor Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package bufferv2

import (
	"reflect"
	"unsafe"
)

// BasePtr returns a pointer to the view's chunk.
func (v *View) BasePtr() *byte {
	hdr := (*reflect.SliceHeader)(unsafe.Pointer(&v.chunk.data))
	return (*byte)(unsafe.Pointer(hdr.Data))
}
