// Copyright 2019 The gVisor Authors.
//
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Package sync provides synchronization primitives.
//
// +checkalignedignore
package sync
